#define __AVR_ATmega328P__

/*
 * MetronomeV1
 * Kerry D. Wong
 * http://www.kerrywong.com
 * 
 * 6/2012
 */
#include <Arduino.h>

// <editor-fold defaultstate="collapsed" desc="constants"> 
const byte BTN_TEMPO_INC = 0; // button index for the tempo increment button
const byte BTN_TEMPO_DEC = 1; // button index for the tempo decrease button

const byte PIN_BTN_TEMPO_INC = 2;
const byte PIN_BTN_TEMPO_DEC = 3;

const byte BUTTONS[] = {PIN_BTN_TEMPO_INC, PIN_BTN_TEMPO_DEC};

const byte PIN_LED_RED = 4;
const byte PIN_LED_GREEN = 5;

const byte PIN_CLK = 6;
const byte PIN_LATCH = 7;
const byte PIN_OE = 8;
const byte PIN_DS = 9;

const byte PIN_DIG_LED = 12;
const byte PIN_SPK = 13;

const byte TEMPO[] = {
    0, //this entry indicates the option for tone A
    40, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 63, 66, 69, 72, 76, 80, 84, 88, 92, 96,
    100, 104, 108, 112, 116, 120, 126, 132, 138, 144, 152, 160, 168, 176, 184, 192
};

//The display range
const byte TEMPO_DISPLAY_RANGE = 37;

//debouncing delay in milliseconds
const long debounceDelay = 50;

//the button event firing interval when the button is kept depressed
//this interval is in terms of the number of main loop iterations 
//recorded by repeatCounter.
const long repeatInterval = 500;

//the length of each beat (both via the red LED and the speaker) in
//terms of the number of main loop iterations recorded by beatDurationCount.
const long beatDuration = 100;

//7seg bitmap:
//A B C D E F G
//7 3 5 4 6 2 1
//bit order LSB --> MSB
//0: segment lit
const byte DIGIT[] = {
    B1000000, //0
    B1101011, //1
    B0100100, //2
    B0100010, //3
    B0001011, //4
    B0010010, //5
    B0010000, //6
    B1101010, //7
    B0000000, //8
    B0000010, //9
};

//bitmap for "-"
const byte DASH = B0111111;

const int A_FREQ = 440;
const int TEMPO_FREQ = 2000; //frequency used to generate the tempo beat
// </editor-fold> 

//<editor-fold defaultstate="collapsed" desc="global variables"> 
//The current selected tempo
byte TempoIndex = 0;

//which digit on the 7seg display is currently being updated
byte currentDigit = 0;
//holds the bitmap for the current displayed numbers
byte currentDisp[2];
//the digit to be displayed.
//the MSB dictates which 7seg the number is displayed on.
byte dispDigit = 0;

byte currentTempo = 0; //the current tempo a*100 + b*10 + c
byte currentTempo0 = 0; //c
byte currentTempo1 = 0; //b
byte currentTempo2 = 0; //a

//the interval in microseconds for a given tempo
unsigned long tempoInterval = 0;

//records the current button states
byte currBtnStates[] = {HIGH, HIGH};

//records the previous button states
byte lastBtnStates[] = {HIGH, HIGH};

//holds the result from digitalRead
byte btnReadings[2];

//holds the debouncing timestamps for each button
long lastTimeStamps[2];

//the counter used to determine button firing intervals
unsigned long repeatCounter = 0;

//the time stamp at the begining of each beat.
unsigned long lastBeatTimeStamp = 0;

//the counter used to determine the length of the on time for the LED and speaker.
unsigned long beatDurationCount = 0;

//whether the led for the tempo beat indicator is on.
bool signalLedOn = false;

//whether any button is pressed.
//if buttonPressed = 0, at least one button is pressed.
int buttonPressed = 1;
// </editor-fold> 

/**
 * setup pins
 */
void setup() {
    pinMode(PIN_SPK, OUTPUT);
    pinMode(PIN_BTN_TEMPO_INC, INPUT);
    digitalWrite(PIN_BTN_TEMPO_INC, HIGH);
    pinMode(PIN_BTN_TEMPO_DEC, INPUT);
    digitalWrite(PIN_BTN_TEMPO_DEC, HIGH);

    pinMode(PIN_DIG_LED, OUTPUT);

    pinMode(PIN_LED_GREEN, OUTPUT);
    pinMode(PIN_LED_RED, OUTPUT);
    digitalWrite(PIN_LED_RED, HIGH);
    digitalWrite(PIN_LED_GREEN, HIGH);
    digitalWrite(PIN_LED_GREEN, LOW); //green on

    pinMode(PIN_CLK, OUTPUT);
    pinMode(PIN_LATCH, OUTPUT);
    pinMode(PIN_OE, OUTPUT);
    pinMode(PIN_DS, OUTPUT);

    //by default, the metronome powers up into calibration mode
    //by playing "A". The 7seg displays "--".
    currentDisp[0] = DASH;
    currentDisp[1] = DASH;

    tone(PIN_SPK, A_FREQ);
    digitalWrite(PIN_OE, LOW); //enable 74H595 outputbeatDurationCount
}


/**
 * check button states and record the states into lastBtnStates[]
 * @return 0 if any button is pressed.
 *         1 if no button is pressed.
 */
int checkButtonStates() {
    int r = 1;

    for (int i = 0; i < 2; i++) {
        btnReadings[i] = digitalRead(BUTTONS[i]);

        if (btnReadings[i] != lastBtnStates[i]) lastTimeStamps[i] = millis();
        if (millis() - lastTimeStamps[i] > debounceDelay) currBtnStates[i] = btnReadings[i];
        lastBtnStates[i] = btnReadings[i];

        if (currBtnStates[i] == 0) r = 0;
    }

    return r;
}

/**
 * main loop
 */
void loop() {
    //determines which 7seg digit to update, each loop updates one of the two
    //7seg displays.
    currentDigit = currentDigit == 0 ? 1 : 0;
    
    //the MSB determines which 7seg digit is displayed.
    dispDigit = currentDigit << 7 | currentDisp[currentDigit];

    buttonPressed = checkButtonStates();
    
    //the repeat counter is used to determine how long does it take to repeat the
    //key pressed action when a key is pressed.
    repeatCounter++;

    //if key is pressed and we are at the desired repeat interval
    if ((repeatCounter % repeatInterval == 0) && (buttonPressed == 0)) {
        if (lastBtnStates[BTN_TEMPO_DEC] == LOW) { //tempo down
            if (TempoIndex > 0) TempoIndex--;
        } else if (lastBtnStates[BTN_TEMPO_INC] == 0) { //tempo up
            if (TempoIndex < TEMPO_DISPLAY_RANGE) TempoIndex++;
        }

        currentTempo = TEMPO[TempoIndex]; //records the selected tempo

        if (currentTempo > 0) {
            noTone(PIN_SPK);
            digitalWrite(PIN_LED_GREEN, HIGH);
            
            //calculate tempo duration in microseconds
            tempoInterval = 60000000l / currentTempo; 

            //currentTempo = currentTempo2 * 100 + currentTemo1 * 10 + currentTempo0
            currentTempo2 = currentTempo / 100;

            //tempo above 100 is indicated by lighting the LED on PIN_DIG_LED pin.
            if (currentTempo2 > 0) {
                digitalWrite(PIN_DIG_LED, HIGH);
                currentTempo -= 100;
            } else {
                digitalWrite(PIN_DIG_LED, LOW);
            }

            currentTempo1 = currentTempo / 10;
            currentDisp[0] = DIGIT[currentTempo1];

            if (currentTempo1 > 0) currentTempo0 = currentTempo - 10 * currentTempo1;
            else currentTempo0 = currentTempo1;

            currentDisp[1] = DIGIT[currentTempo0];
            
            lastBeatTimeStamp = micros();
            beatDurationCount = 0;
        } else {// if the selected tempo is 0, we are in calibration mode
            digitalWrite(PIN_LED_GREEN, LOW);
            currentDisp[0] = DASH;
            currentDisp[1] = DASH;
            tone(PIN_SPK, A_FREQ);
        }
    }

    if (currentTempo > 0) {
        beatDurationCount++;

        // if the beat interval passed, we initiate a new beat
        if (micros() - lastBeatTimeStamp >= tempoInterval) { 
            lastBeatTimeStamp = micros();
            beatDurationCount = 0;

            //the beat (LED and speaker) is on only when the beat duration has not passed
            //the beat duration is less then the beat interval
            if ((beatDurationCount < beatDuration) && (!signalLedOn)) {
                digitalWrite(PIN_LED_RED, LOW);
                tone(PIN_SPK, TEMPO_FREQ);
                signalLedOn = true;
            }
        } else { //if the beat interval passed, we turn off the speaker and dim the LED.
            if (beatDurationCount > beatDuration) {
                digitalWrite(PIN_LED_RED, HIGH);
                signalLedOn = false;
                noTone(PIN_SPK);
                beatDurationCount = 0;
            }
        }
    }

    //output a single digit (determined by the MSB in dispDigit)
    digitalWrite(PIN_LATCH, LOW);
    shiftOut(PIN_DS, PIN_CLK, MSBFIRST, dispDigit);
    digitalWrite(PIN_LATCH, HIGH);
}
